import finufft as fn
from numpy import *
from numpy.fft import *
from numpy.linalg import norm
from matplotlib.pyplot import *
import matplotlib.ticker as ticker

nPix = 256
goldrat = (1+sqrt(5))/2
sWind = "poly" # "cos" "poly" "es"

fft = lambda x: fftshift(fftn(ifftshift(x)))
ift = lambda x: fftshift(ifftn(ifftshift(x)))
nufft = lambda k, x: fn.nufft1d2((2*pi)*k, x, eps=1e-8)
nuift = lambda k, x, n: fn.nufft1d1((2*pi)*k, x, n, eps=1e-8)

def winfun(x, p, sWind):
    if sWind=="cos": arrW = cos(pi/2 * x)**p
    elif sWind=="poly": arrW = 1 - (abs(x))**p
    elif sWind=="es": arrW = exp(p*sqrt(1-x*x))/exp(p)
    else: raise NotImplementedError()
    return arrW

arrP = arange(1.0,3.0,0.1) # arange(1.80, 2.00, 0.01)
arrErr = None
for iTest in range(100):
    print(f"iTest: {iTest}")
    random.seed(iTest)
    # arrK = random.randn(int(sqrt(2*pi)*e**4.5*nPix/6))/6
    '''
    PDF of 3sigma is `1/sqrt(2*pi)/e**4.5`, possibility to drop in `6/nPix` region at `3sigma` is `6/sqrt(2*pi)/e**4.5/nPix`
    '''
    arrK = random.randn(int(sqrt(2*pi)*e**2*nPix/4))/4
    '''
    PDF of 2sigma is `1/sqrt(2*pi)/e**2`, possibility to drop in `4/nPix` region at `2sigma` is `4/sqrt(2*pi)/e**2/nPix`
    '''
    # arrK = random.randn(int(sqrt(2*pi*e)*nPix/2))/2
    '''
    PDF of 1sigma is `1/sqrt(2*pi*e)`, possibility to drop in `2/nPix` region at `1sigma` is `2/sqrt(2*pi*e)/nPix`
    '''
    # figure()
    # hist(arrK, bins=100)
    # savefig("fig.pdf", dpi=300)
    # exit()
    # arrK = random.rand(2*nPix)-0.5
    # arrK = (iTest/nPix + arange(nPix*goldrat)/goldrat)%1-0.5
    arrK[0] = 0
    arrK = arrK[abs(arrK)<0.5]
    print(f"Nk: {arrK.size}, range: {arrK.min():.3f}, {arrK.max():.3f}")

    arrDcf = ones_like(arrK, dtype=complex128) # E
    # arrDcf = abs(arrK)+1/nPix + 0j # E
    arrDcf /= arrDcf.sum()
    arrPsf = nuift(arrK, arrDcf, (2*nPix)-1)/size(arrK) # P

    lstErr = []
    for iP in range(arrP.size):
        p = arrP[iP]
        x = linspace(-nPix,nPix,2*nPix-1)
        arrW = winfun(x/nPix, p, sWind)
        
        arrPsfStar = nuift(arrK, arrDcf/nufft(arrK, arrPsf*arrW), (2*nPix)-1)/size(arrK)
        arrWeightPsf = arrPsfStar * log2(1 + abs(linspace(-(nPix-1),nPix-1,2*nPix-1,1))) / abs(arrPsfStar[nPix-1])
        
        lstErr.append(norm(arrWeightPsf)) # , ord=inf (min-max)
    
    if arrErr is None: arrErr = array(lstErr)
    else: arrErr = maximum(arrErr, array(lstErr))
    
arrErr[isnan(arrErr)] = max(arrErr[~isnan(arrErr)])
iPOpt = argmin(arrErr)
    
# fig = figure(figsize=(5,3), dpi=150)
# ax = fig.add_subplot(111)
# lstC = ["tab:blue"]*arrP.size
# lstC[iPOpt] = "tab:red"
# ax.scatter(arrP, arrQuaErr, c=lstC)
# ax.set_xlabel(r"$p$")
# ax.set_ylabel(r"$\|P^\star_\mathrm{out}(\mathbf{x})\|$")
# ax.set_ylim(0,10)
# ax.text(arrP[iPOpt], arrQuaErr[iPOpt]-0.1, rf"$p={arrP[iPOpt]:.1f}$"+"\n"+rf"$\|P^\star_\mathrm{{out}}(\mathbf{x})\|={arrQuaErr[iPOpt]:.3f}$", horizontalalignment='center', verticalalignment='top')

wFPage = 4.77
fig = figure(figsize=(wFPage, wFPage*3/3), dpi=300)

ax = fig.add_subplot(211)
lstC = ["tab:blue"]*arrP.size
lstC[iPOpt] = "tab:red"
ax.scatter(arrP, arrErr, c=lstC)
ax.set_ylim(2,10) # arrErr.min()*0.9,arrErr.min()*1.1)
ax.set_xlabel(r"$p$")
ax.set_ylabel(r"$\|P^\star_\mathrm{out}(\mathbf{x})\|$")
ax.text(arrP[iPOpt], arrErr[iPOpt]-0.1, rf"$p={arrP[iPOpt]:.1f}$"+"\n"+rf"$\|P^\star_\mathrm{{out}}(\mathbf{{x}})\|={arrErr[iPOpt]:.3f}$", horizontalalignment='center', verticalalignment='top')
ax.set_title("(A)", loc="left")

ax = fig.add_subplot(212)
x = linspace(0,1,1000)
y = winfun(x, arrP[iPOpt], sWind)

ax.plot(x, y, "-")
ax.set_xlabel(r"$\|\mathbf{\bar{x}}\|$")
ax.set_ylabel(r"$W^\star(\mathbf{x})$")
if sWind=="cos": sFunc = rf"$W^\star(\mathbf{{x}})=\cos(\pi\mathbf{{\bar{{x}}}}/2)^{{{arrP[iPOpt]:.1f}}}$"
elif sWind=="poly": sFunc = rf"$W^\star(\mathbf{{x}})=1-\|\mathbf{{\bar{{x}}}}\|^{{{arrP[iPOpt]:.1f}}}$"
else: sFunc = rf"$W^\star(\mathbf{{x}})$"
ax.text(1,y[0], sFunc, horizontalalignment='right', verticalalignment='top')
ax.set_title("(B)", loc="left")

fig.subplots_adjust(0.15,0.1,0.99,0.95, wspace=0.0, hspace=0.5)
fig.savefig("temp/figure/PSearch.png", dpi=300)

show()


from numpy import *
from scipy.ndimage import gaussian_filter
from .Type import Part

def genPhMap(nDim:int=2, nPix:int=256, mean:int|float|None=None, std:int|float=pi/16) -> ndarray:
    """
    # return
    smooth complex rotation factor
    """
    if mean == None: mean = random.uniform(-pi,pi)
    mapPh = random.uniform(-pi, pi, [nPix for _ in range(nDim)])
    sigma = nPix/4
    mapPh = gaussian_filter(mapPh, sigma)
    # normalize
    mapPh -= mapPh.mean(); mapPh = asarray(mapPh)
    mapPh /= mapPh.std()
    mapPh *= std
    mapPh += mean
    # convert to rotation factor
    mapPh = exp(1j*mapPh)
    mapPh = mapPh/abs(mapPh)
    return mapPh

def genB0Map(nDim:int=2, nPix:int=256, mean:int|float=0, std:int|float=1e-6*(2*pi*42.58e6*3)) -> ndarray:
    """
    # return
    smooth random number
    """
    mapB0 = random.uniform(-1, 1, [nPix for _ in range(nDim)])
    sigma = nPix/4
    mapB0 = gaussian_filter(mapB0, sigma)
    # normalize
    mapB0 -= mapB0.mean(); mapB0 = asarray(mapB0)
    mapB0 /= mapB0.std()
    mapB0 *= std
    mapB0 += mean
    return mapB0

def genCsm(nDim:int=2, nPix:int=256, nCh:int=12, mean:int|float|None=None, std:int|float=pi/16) -> ndarray:
    if mean == None: mean = random.uniform(-pi,pi)
    mapC = zeros([nCh,*(nPix for _ in range(nDim))], dtype=complex128)
    arrCoor = meshgrid\
    (
        *(linspace(-0.5,0.5,nPix,0) for _ in range(nDim)),
        indexing="ij"
    ); arrCoor = array(arrCoor).transpose(*arange(1,nDim+1), 0)
    arrTht = linspace(0,2*pi,nCh,0)
    arrCoorCoil = zeros([nCh,nDim], dtype=float64)
    arrCoorCoil[:,-2:] = 1*array([sin(arrTht), cos(arrTht)]).T
    if nDim == 3:
        arrCoorCoil[0::2,0] = 0.2
        arrCoorCoil[1::2,0] = -0.2
    for iCh in range(nCh):
        mapC[iCh] = genPhMap(nDim=nDim, nPix=nPix, mean=mean, std=std)
        dist = sqrt(sum((arrCoor - arrCoorCoil[iCh])**2, axis=-1))
        mapC[iCh] *= exp(-dist)
    return mapC

def genAmp(tScan:int|float, tRes:int|float, cyc:int|float, isRand:bool=True):
    """
    # parameter
    `tScan`: how many seconds this waveform contains
    `tRes`: how many ticks per second
    `cyc`: cycle of desired signal in second
    `isRand`: whether to randomize the waveform
    """
    nT = tScan*tRes

    if isRand:
        arrT = sort(random.rand(nT)*tScan)
        arrAmp = sin(2*pi/cyc*arrT)

        sigma = cyc*tRes/8
        arrAmp = gaussian_filter(arrAmp, sigma)
    else:
        arrT = linspace(0, tScan, nT)
        arrAmp = sin(2*pi/cyc*arrT)

    return arrAmp

def Enum2M0(arrPhan:ndarray) -> ndarray:
    mapM0 = zeros_like(arrPhan, dtype=float64)
    mapM0[arrPhan==Part.Air.value] = 0 # random.randn(sum(arrPhan==Part.Air.value))*1e-3
    mapM0[arrPhan==Part.Fat.value] = 1.0
    mapM0[arrPhan==Part.Body.value] = 0.5
    mapM0[arrPhan==Part.Myo.value] = 0.2
    mapM0[arrPhan==Part.Blood.value] = 0.8
    mapM0[arrPhan==Part.Other.value] = 1.0
    return mapM0

def Enum2T1(arrPhan:ndarray) -> ndarray:
    mapT1 = zeros_like(arrPhan, dtype=float64)
    mapT1[arrPhan==Part.Air.value] = inf # 10000e-3 #random.uniform(1e-3, 10000e-3, sum(arrPhan==Part.Air.value))
    mapT1[arrPhan==Part.Fat.value] = 350e-3
    mapT1[arrPhan==Part.Body.value] = 1600e-3
    mapT1[arrPhan==Part.Myo.value] = 1300e-3
    mapT1[arrPhan==Part.Blood.value] = 1500e-3
    mapT1[arrPhan==Part.Other.value] = 1000e-3 # arbitary value
    return mapT1

def Enum2T2(arrPhan:ndarray) -> ndarray:
    mapT2 = zeros_like(arrPhan, dtype=float64)
    mapT2[arrPhan==Part.Air.value] = 1e-6 # random.uniform(1e-3, 1000e-3, sum(arrPhan==Part.Air.value))
    mapT2[arrPhan==Part.Fat.value] = 75e-3
    mapT2[arrPhan==Part.Body.value] = 40e-3
    mapT2[arrPhan==Part.Myo.value] = 50e-3
    mapT2[arrPhan==Part.Blood.value] = 225e-3
    mapT2[arrPhan==Part.Other.value] = 10e-3 # arbitary value
    return mapT2

def Enum2Om(arrPhan:ndarray, B0:int|float=3) -> ndarray:
    mapOm = zeros_like(arrPhan, dtype=float64)
    ppm2om = 1e-6*(2*pi*42.58e6*B0)
    mapOm[arrPhan==Part.Air.value] = 0 # random.uniform(1e-3, 10e-3, sum(arrPhan==Part.Air.value))*ppm2om
    mapOm[arrPhan==Part.Fat.value] = 3.5*ppm2om
    mapOm[arrPhan==Part.Body.value] = 0
    mapOm[arrPhan==Part.Myo.value] = 0
    mapOm[arrPhan==Part.Blood.value] = 0
    mapOm[arrPhan==Part.Other.value] = 3.5*ppm2om
    return mapOm


import h5py
import numpy as np
import matplotlib.pyplot as plt
# Import svds (Singular Value Decomposition for sparse/truncated calculation)
from scipy.sparse.linalg import svds
from sigpy.mri.app import EspiritCalib

# --- Configuration ---
pathRes = "/mnt/d/LProject/DcfBenchmark/resource/"
# dsid = "1b197efe-9865-43be-ac24-f237c380513e"
dsid = "2588bfa8-0c97-478c-aa5a-487cc88a590d"
FILE_PATH = f'{pathRes[:-1]}/{dsid}.h5'
DATASET_PATH = '/dataset/data'
TARGET_NX = 320 # Target readout size after cropping (to remove oversampling)

with h5py.File(FILE_PATH, 'r') as f:
    print(f"Loading data from {DATASET_PATH}...")
    
    # Load the structured array (includes 'head' and 'data')
    dataset = f[DATASET_PATH][:]
    
    # Extract the k-space data field (array of arrays) and header field
    data = dataset['data']
    head = dataset['head'] 
    print("kspace_data_field.dtype", data.dtype)
    
    if len(data) == 0:
        raise ValueError("No k-space data found.")
    
    # --- Determine 3D Dimensions from metadata ---
    
    print("kspace_data_field[0].shape:", data[0].shape)
    print("kspace_data_field[0].dtype:", data[0].dtype)
    
    nCh = head[0]['active_channels']
    nRO = data[0].shape[0]
    
    # Calculate N_samples_readout (Nx_raw)
    nX = nRO // nCh // 2
    
    nAcq = len(data)
    
    # Infer the phase encoding steps (Ny and Nz) using the kspace indices
    vecIdxPhEnc1 = head['idx']['kspace_encode_step_1']
    vecIdxPhEnc2 = head['idx']['kspace_encode_step_2'] # Z dimension candidate 1 (Kz)
    vecIdxSlice = head['idx']['slice'] # Z dimension candidate 2 (Slice index)
    
    # Debugging output for indices
    print(f"DEBUG: vecIdxPhEnc1 range: {vecIdxPhEnc1.min()} to {vecIdxPhEnc1.max()}")
    print(f"DEBUG: vecIdxPhEnc2 range: {vecIdxPhEnc2.min()} to {vecIdxPhEnc2.max()}")
    print(f"DEBUG: vecIdxSlice range: {vecIdxSlice.min()} to {vecIdxSlice.max()}")
    
    nY = vecIdxPhEnc1.max() + 1 # Ny
    
    # nZ = vecIdxPhEnc2.max() + 1
    nZ = vecIdxSlice.max() + 1
    
    print(f"Inferred Dimensions: Channels={nCh}, Readout Raw (Nx_raw)={nX}, Phase Y (Ny)={nY}, Phase Z (Nz)={nZ}")
    
    # --- Robust Reshaping of K-space Data (Mapping using 3D indices) ---
    
    # Initialize the final k-space grid: [N_channels, Nz, Ny, Nx_raw]
    kspMulCh = np.zeros((nCh, nZ, nY, nX), dtype=np.complex64)
    
    # Iterate through all acquisitions and place them at their correct 3D index
    for i in range(nAcq):
        iY = vecIdxPhEnc1[i]
        iZ = vecIdxSlice[i] # Use the determined Z index array
        arrS = data[i]
        
        # Place data into the 4D grid: [Channel, Nz, Ny, nX]
        kspMulCh[:, iZ, iY, :] = (arrS[0::2] + arrS[1::2]*1j).reshape(nCh, nX)
    kspMulCh = np.fft.fftn(kspMulCh, axes=(-3,))
    kspMulCh = np.fft.fftshift(kspMulCh, axes=(-3,))
    kspMulCh = np.fft.ifftn(kspMulCh, axes=(-3,))
    
    print(f"K-space array shape for reconstruction: {kspMulCh.shape}")
    
    # Ecalib
    arrCsm = EspiritCalib(kspMulCh)
    print(arrCsm.shape)
    exit()
    plt.figure()
    plt.imshow()
    

    # --- Step 1: 3D Inverse FFT (iFFT) ---
    
    tupAxes = (-2,-1)
    _ = np.fft.ifftshift(kspMulCh, axes=tupAxes)
    _ = np.fft.ifftn(_, axes=tupAxes)
    imgMulCh = np.fft.fftshift(_, axes=tupAxes)
    
    iY0 = nY//2-nZ//2; iY1 = iY0 + nZ
    iX0 = nX//2-nZ//2; iX1 = iX0 + nZ
    imgMulCh = imgMulCh[:,:,iY0:iY1,iX0:iX1] # crop
    
    print(f"Full Image domain array shape: {imgMulCh.shape}")

    # --- Step 1b: Crop Readout Oversampling (nX) ---
    
    nCh, nZ, nY, nX = imgMulCh.shape
    
    # --- Step 2: SVD Channel Combination using svds (3D) ---
    
    imgMulCh = imgMulCh.transpose(0,3,2,1)[:,::-1,:,:]
    
    
    # Reshape back to 3D image volume: [Nz, Ny, nX]
    imgRec = imgFlat.reshape(nZ, nY, nX)
    imgRec /= np.percentile(np.abs(imgRec), 95)
    np.save(f"{pathRes[:-1]}/{dsid}.npy", imgRec) # save
    
    print(f"Final combined image volume shape: {imgRec.shape}")
    
    # --- Step 3: Display Center Slices ---
    
    # Extract slices
    sliceZ0 = imgRec[nZ//2, :, :] # Z fixed, showing Y-X plane
    sliceY0 = imgRec[:, nY//2, :] # Y fixed, showing Z-X plane
    sliceX0 = imgRec[:, :, nX//2] # X fixed, showing Z-Y plane
    
    # Display setup
    fig, axes = plt.subplots(2, 3, figsize=(15, 5))
    axes = np.ravel(axes)
    
    # 1. Axial (Z plane, typically the acquisition plane)
    axes[0].imshow(np.abs(sliceZ0), origin="lower", cmap='gray')
    axes[0].set_title(f'Axial Slice (Z={nZ//2}) | Y x X ({nY}x{nX})')
    axes[0].axis('off')
    
    # 2. Coronal (Y plane)
    axes[1].imshow(np.abs(sliceY0), origin="lower", cmap='gray')
    axes[1].set_title(f'Coronal Slice (Y={nY//2}) | Z x X ({nZ}x{nX})')
    axes[1].axis('off')
    
    # 3. Sagittal (X plane)
    axes[2].imshow(np.abs(sliceX0), origin="lower", cmap='gray')
    axes[2].set_title(f'Sagittal Slice (X={nX//2}) | Z x Y ({nZ}x{nY})')
    axes[2].axis('off')
    
    # 4. Axial (Z plane, typically the acquisition plane)
    axes[3].imshow(np.angle(sliceZ0), origin="lower", cmap='hsv', vmin=-np.pi, vmax=np.pi)
    axes[3].set_title(f'Axial Slice (Z={nZ//2}) | Y x X ({nY}x{nX})')
    axes[3].axis('off')
    
    # 5. Coronal (Y plane)
    axes[4].imshow(np.angle(sliceY0), origin="lower", cmap='hsv', vmin=-np.pi, vmax=np.pi)
    axes[4].set_title(f'Coronal Slice (Y={nY//2}) | Z x X ({nZ}x{nX})')
    axes[4].axis('off')
    
    # 6. Sagittal (X plane)
    axes[5].imshow(np.angle(sliceX0), origin="lower", cmap='hsv', vmin=-np.pi, vmax=np.pi)
    axes[5].set_title(f'Sagittal Slice (X={nX//2}) | Z x Y ({nZ}x{nY})')
    axes[5].axis('off')
    
    fig.suptitle('3D Reconstructed MRI Volume (iFFT + SVD Combination)')
    plt.show()


from setuptools import setup

_packages = \
[
    "mrarbdcf",
]

_package_dir = \
{
    "mrarbdcf":"./mrarbdcf_src/", 
}

setup\
(
    name = 'mrarbdcf',
    packages = _packages,
    package_dir = _package_dir,
    include_package_data = True
)
